package net.sf.corn.converter;

import java.util.Collections;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * Abstract base class for object converters
 * @author Serhat Dirik
 */
public abstract class AbstractConverter<E> implements IConverter<E> {
	/**
	 * list of the supported java classes for the conversion
	 */
	protected List<Class<?>> convertibles = new CopyOnWriteArrayList<Class<?>>();
	
	private final static String PRIMITIVE_BOOLEAN="boolean"; 
	private final static String PRIMITIVE_CHAR="char";
	private final static String PRIMITIVE_INT="int";
	private final static String PRIMITIVE_DOUBLE="double";
	private final static String PRIMITIVE_FLOAT="float";
	private final static String PRIMITIVE_BYTE="byte";
	private final static String PRIMITIVE_SHORT="short";
	private final static String PRIMITIVE_LONG="long";
	/**
	 * List of the convertible java types by this class
	 * @return list of conversion supported classes
	 */
	protected List<Class<?>> getConvertibles() {
		return convertibles;
	}
	/**
	 * Adds the given java class to the convertibles list
	 * @param convertible a java class that its conversion supported 
	 * @return true if its added successfully
	 */
	protected boolean addConvertible(Class<?> convertible){
		return convertibles.add(convertible);
	}

	public List<Class<?>> findConvertibles() {
		return Collections.unmodifiableList(convertibles);
	}
	/**
	 * Tells whether the given class conversion is supported by this class 
	 * @param clazz the class to be checked
	 * @return true if it's supported
	 */
	protected boolean checkIfConvertibleFromRegisteredTypes(Class<?> clazz) {
		for(Class<?> type:convertibles){
			if ((clazz == null && !type.isPrimitive()) || (clazz != null && type.isAssignableFrom(clazz)) ) return true;
		}
		return false;
	}
	
	
	public boolean isConvertible(String className) {
		if (className == null || className.trim().length() == 0) {
			return true;
		}
		Class<?> clazz;
		if(className.trim().equals(PRIMITIVE_BOOLEAN)){
			clazz = boolean.class;
		}else if(className.trim().equals(PRIMITIVE_BYTE)){
			clazz = byte.class;
		}else if(className.trim().equals(PRIMITIVE_CHAR)){
			clazz = char.class;
		}else if(className.trim().equals(PRIMITIVE_DOUBLE)){
			clazz = double.class;
		}else if(className.trim().equals(PRIMITIVE_FLOAT)){
			clazz = float.class;
		}else if(className.trim().equals(PRIMITIVE_INT)){
			clazz = int.class;
		}else if(className.trim().equals(PRIMITIVE_LONG)){
			clazz = long.class;
		}else if(className.trim().equals(PRIMITIVE_SHORT)){
			clazz = short.class;
		}else {
			try {
				clazz = Class.forName(className);
			} catch (ClassNotFoundException e) {
				return false;
			}
		}
		
		return isConvertible(clazz);
	}
	
	public boolean isConvertible(Class<?> clazz) {
		return checkIfConvertibleFromRegisteredTypes(clazz);
	}
		
	public final Object toJava(E obj) throws ConversionException {
		return toJava(null, obj);
	}

	public final Object toJava(Class<?> origClazz, E origObj)
			throws ConversionException {
		if (origObj == null || isNullType(origObj)) { 
			return null;
		}
		// TODO Consider early evaluation of isEmpty
		if (!isConvertible(origClazz)) {
			throw new ConversionException(this.getClass().getName()
					+ " can't be used as converter for " + origClazz.getName());
		}
		try {
			return inToJava(origClazz, origObj);
		} catch (ConversionException e) {
			throw e;
		} catch (Exception e) {
			throw new ConversionException(this.getClass().getName()
					+ " can't be used as converter for "
					+ origObj.getClass().getName(), e);
		}
	}

	public final E toTargetType(Object o) throws ConversionException {
		return toTargetType(o, null);
	}

	public final E toTargetType(Object o, ObjectReferenceMap referenceMap)
			throws ConversionException {
		if (o == null) {
			return getNullType();
		}
		if (!isConvertible(o.getClass())) {
			throw new ConversionException(this.getClass().getName()
					+ " can't be used as converter for " + o.getClass());
		}
		if (isSubjectForCircularReferenceCheck()) {
			if (referenceMap == null)
				referenceMap = new ObjectReferenceMap();
			referenceMap.putObject(o);
		}
		E ret = inToTarget(o, referenceMap);
		if (isSubjectForCircularReferenceCheck() && referenceMap != null) {
			referenceMap.removeObject(o);
		}
		return ret;
	}
	/**
	 * Checks if the given targeted type object represents the null value  
	 * @param obj the object to be checked
	 * @return true if the given object is null
	 */
	protected abstract boolean isNullType(E obj);
	/**
	 * Gives the null represents targeted typed object
	 * @return null represents object
	 */
	protected abstract E getNullType(); 
	/**
	 * Tells if the conversion supported classes are subject for circular object reference
	 * @return true if supported types are subject to the circular references
	 */
	protected abstract boolean isSubjectForCircularReferenceCheck();
	/**
	 * Inner conversion method to target type
	 * @param o object to be converted 
	 * @param referenceMap circular reference map
	 * @return converted object
	 * @throws ConversionException thrown if any error occurs during conversion
	 */
	protected abstract E 	  inToTarget(Object o,	ObjectReferenceMap referenceMap) throws ConversionException;
	/**
	 * Inner conversion method to java type
	 * @param clazz java class as target to this conversion
	 * @param obj object to be converted
	 * @return java object
	 * @throws ConversionException thrown if any error occurs during conversion
	 */
	protected abstract Object inToJava(Class<?> clazz, E obj)		throws ConversionException;

}
